﻿using System;
using System.IO;
using System.Net.Sockets;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Diagnostics;
using System.Net;

namespace locationserver
{
    class Program
    {        
        public static LeaderboardItem[] leaderboard = new LeaderboardItem[5];

        public static void runServer()
        {
            TcpListener listener;
            
            listener = new TcpListener(IPAddress.Any, 43);
            listener.Start();

            while (true)
            {
                Socket connection = listener.AcceptSocket();
                NetworkStream socketStream = new NetworkStream(connection);
                socketStream.ReadTimeout = 1000;
                try
                {
                    doRequest(socketStream);
                }
                catch (Exception e)
                {
                    Debug.Print(e.ToString());
                }
                finally
                {
                    socketStream.Close();
                    connection.Close();
                }
            }
        }

        public static void doRequest(NetworkStream connection)
        {
            StreamWriter sw = new StreamWriter(connection);
            StreamReader sr = new StreamReader(connection);

            ProtocolChoice thing = new ProtocolChoice();

            string inputStreamString = sr.ReadLine();           //Input String
            string inputStreamWhole = sr.ReadToEnd();           //Read the whole message
            char[] deliminators = new char[1];
            deliminators[0] = ' ';
            string[] splitString = inputStreamString.Split(deliminators); //This takes in arguments by splitting them using a defined character. In this case a space between arguments.

            if (splitString.Length == 1)    
            {
                //SEND THE WHOLE LEADERBOARD       
                Console.WriteLine(splitString[0]);
                thing.SendWholeLeaderboard(sw, ref leaderboard);
                sw.Flush();
            }
            else if (splitString.Length == 2)//Changes it and then tells the client OK.
            {
                //Put name and score into leaderboard
                Console.WriteLine("Inserting {0} with a score of {1} into the leaderboard", splitString[0], splitString[1]);
                thing.SingleUpdate(splitString, sw, sr, ref leaderboard);
                for (int i = 0; i < leaderboard.Length; i++)
                {
                    Console.WriteLine(leaderboard[i].NAME + " - " + leaderboard[i].SCORE);
                }
                sw.WriteLine("OK");
                sw.Flush();
                SaveLeaderboard("leaderboard.txt");
            }
            else if (splitString.Length > 2)
            {
                //Update the leaderboard with all the things
                
                thing.WholeLeaderBoardUpdate(splitString, sw, sr, ref leaderboard);
                for (int i = 0; i < leaderboard.Length; i++)
                {
                    Console.WriteLine(leaderboard[i].NAME + " - " + leaderboard[i].SCORE);
                }
                SaveLeaderboard("leaderboard.txt");
            }

        }

        #region Loading Highscore
        public static void LoadLeaderboard(System.IO.TextReader textIn)
        {
            leaderboard[0] = new LeaderboardItem(textIn.ReadLine(), int.Parse(textIn.ReadLine()));
            leaderboard[1] = new LeaderboardItem(textIn.ReadLine(), int.Parse(textIn.ReadLine()));
            leaderboard[2] = new LeaderboardItem(textIn.ReadLine(), int.Parse(textIn.ReadLine()));
            leaderboard[3] = new LeaderboardItem(textIn.ReadLine(), int.Parse(textIn.ReadLine()));
            leaderboard[4] = new LeaderboardItem(textIn.ReadLine(), int.Parse(textIn.ReadLine()));
            

            textIn.Close();
        }

        public static void LoadLeaderboard(string filename)
        {
            System.IO.TextReader textIn = null;
            try
            {
                textIn = new System.IO.StreamReader(filename);
                LoadLeaderboard(textIn);
            }
            catch (Exception e)
            {
                throw e;
            }
            finally
            {
                if (textIn != null) textIn.Close();
            }
        }
        #endregion

        #region Saving Highscore
        public static bool SaveLeaderboard(System.IO.TextWriter textOut)
        {
            try
            {
                for (int i = 0; i < 5; i++)
                {
                    textOut.WriteLine(leaderboard[i].NAME);
                    textOut.WriteLine(leaderboard[i].SCORE);
                }
            }
            catch
            {
                return false;
            }
            return true;
        }

        public static void SaveLeaderboard(string filename)
        {
            System.IO.TextWriter textOut = null;
            try
            {
                textOut = new System.IO.StreamWriter(filename);
                SaveLeaderboard(textOut);
            }
            catch (Exception e)
            {
                throw e;
            }
            finally
            {
                if (textOut != null) textOut.Close();
            }
        }
        #endregion

        public static void Init()
        {
            leaderboard[0] = new LeaderboardItem("Adam", 2100);
            leaderboard[1] = new LeaderboardItem("Adam", 900);
            leaderboard[2] = new LeaderboardItem("Adam", 800);
            leaderboard[3] = new LeaderboardItem("Adam", 500);
            leaderboard[4] = new LeaderboardItem("Adam", 300);

            SaveLeaderboard("leaderboard.txt");
        }

        static void Main(string[] args)
        {
            LoadLeaderboard("leaderboard.txt");
            //Init();
            Console.WriteLine("SERVER START: CONTENT LOADED");

            for (int i = 0; i < leaderboard.Length; i++)
            {
                Console.WriteLine(leaderboard[i].NAME + " - " + leaderboard[i].SCORE);
            }

            runServer();
            SaveLeaderboard("leaderboard.txt");
        }
    }
}
